#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <memory.h>
#include <stdlib.h>
#include <math.h>


typedef unsigned char  byte;
typedef unsigned short word;
byte far* VGA_PTR = (byte far*)MK_FP(0xA000,0);
const unsigned int INPUT_STATUS = 0x03DA;
const unsigned int VRETRACE = 0x08;
const unsigned int PAL_ANIM_START = 96;
const unsigned int PAL_ANIM_COUNT = 122;

word ReadClock(void)
{
	return *((word far*)MK_FP(0,0x46C)); // This points to the 18.2hz system clock
}
word ClockToSeconds(word ticks) { return (word)(((unsigned long)ticks)*1000 / 18200); }

struct midiinfo
{
	struct clioptions *params;
	struct trackinfodata *trackinfo;
	struct midi_event_t *eventscache;
	struct playinfo_t *playinfo;
};

extern "C" int bgmidInit(int argc,char **argv,struct clioptions **pparams,struct trackinfodata **ptrackinfo,struct midi_event_t **peventscache,struct playinfo_t**pplayinfo);
extern "C" void bgmidBeginPlay(struct clioptions *params,struct trackinfodata *trackinfo,struct midi_event_t *eventscache,struct playinfo_t *playinfo);
extern "C" int bgmidTick(struct clioptions *params,struct trackinfodata *trackinfo,struct midi_event_t *eventscache,struct playinfo_t *playinfo);
extern "C" void bgmidShutdown(struct clioptions **pparams,struct trackinfodata **ptrackinfo,struct midi_event_t **peventscache,struct playinfo_t**pplayinfo);


struct BITMAP256
{
	word width;
	word height;
	byte palette[256*3];

	BITMAP256() : width(0), height(0) {}

	bool LoadToVGA(char *file)
	{
		// Fill entire palette with black to hide scan loading
		{
			outp(0x03C8,0); // Notify of full palette fill
			for (int i=0;i<256*3;i++)
				outp(0x03C9,0);
		}

		FILE *fp;
		if ((fp = fopen(file,"rb")) == NULL)
		{
			printf("Error opening file %s.\n",file);
			return false;
		}

		if (fgetc(fp)!='B' || fgetc(fp)!='M')
		{
			fclose(fp);
			printf("%s is not a bitmap file.\n",file);
			return false;
		}

		word num_colors;

		// Read in the width and height of the image, and the
		// number of colors used; ignore the rest
		fskip(fp,16);
		fread(&width,sizeof(word),1,fp);
		fskip(fp,2);
		fread(&height,sizeof(word),1,fp);
		fskip(fp,22);
		fread(&num_colors,sizeof(word),1,fp);
		fskip(fp,6);

		if (width > 320 || height > 200)
		{
			fclose(fp);
			printf("%s is too big to fit on screen.\n",file);
			return false;
		}

		// Assume we are working with an 8-bit file
		if (num_colors == 0)
			num_colors = 256;

		// Read the palette information
		byte *pPal = palette;
		for (long index=0;index<num_colors;index++)
		{
			byte b = (byte)(fgetc(fp) >> 2);
			byte g = (byte)(fgetc(fp) >> 2);
			byte r = (byte)(fgetc(fp) >> 2);
			fgetc(fp); // Alpha?
			*(pPal++) = r;
			*(pPal++) = g;
			*(pPal++) = b;
		}

		// Read the bitmap
		byte far *scanline = VGA_PTR + (word)(320*(height-1));
		for (long index=height-1;index>=0;index--)
		{
			byte line[320];
			fread(line,1,width,fp);
			_fmemcpy(scanline,line,width);
			scanline -= (word)320;
		}
		fclose(fp);
		return true;
	}

	/*void FunkyPalette(void)
	{
		srand(ReadClock());
		byte rm = rand()%64;
		byte gm = rand()%64;
		byte bm = rand()%64;
		byte max = rm>gm?rm:gm; max = max>bm?max:bm;
		if (max < 50)
		{
			byte diff = 50-max;
			rm += diff;
			gm += diff;
			bm += diff;
		}
		for (long l=PAL_ANIM_START;l<PAL_ANIM_START+PAL_ANIM_COUNT;l++)
		{
			float v = (l-PAL_ANIM_START)/(float)(PAL_ANIM_COUNT-1);
			v = (float)fabs(sin(v*3.1415f*2.0f))*0.5f+0.5f;
			palette[l*3+0] = (byte)((rm * v));
			palette[l*3+1] = (byte)((gm * v));
			palette[l*3+2] = (byte)((bm * v));
		}
	}*/

private:
	static void fskip(FILE *fp,int num_bytes)
	{
		for (int i=0; i<num_bytes; i++)
			fgetc(fp);
	}
};

/* poll the keyboard, and return the next input key if any, or -1 */
bool AnyKeyPressed(void)
{
  union REGS regs;
  regs.h.ah = 0x0B;
  int86(0x21, &regs, &regs);
  return (regs.h.al == 0xFF);
}

int VerticalRetraceDone(void) { return (inp(INPUT_STATUS) & VRETRACE) == 0; }
int VerticalRefreshDone(void) { return (inp(INPUT_STATUS) & VRETRACE) != 0; }

void WaitForRetrace(void)
{
	// Wait until done with vertical retrace
	while (!VerticalRetraceDone());
	while (!VerticalRefreshDone());
}

void AnimateImagePalette(const BITMAP256& bmp, midiinfo& mid)
{
	bool midPlayDone = false;

	// Black screen to skip some midi
	{
		word startTime = ReadClock();
		while (ClockToSeconds(ReadClock()-startTime) < 3)
		{
			if (!midPlayDone)
				midPlayDone = bgmidTick(mid.params,mid.trackinfo,mid.eventscache,mid.playinfo) > 0;
		}
	}

	// Fill palette
	outp(0x03C8,0); // Notify of full palette fill
	const byte *pal = bmp.palette;
	for (int i=0;i<256;i++)
	{
		outp(0x03C9,*pal++); // R
		outp(0x03C9,*pal++); // G
		outp(0x03C9,*pal++); // B
	}

	// Now rotate palette
	byte pal_rot[PAL_ANIM_COUNT*3];
	memcpy(pal_rot,bmp.palette+PAL_ANIM_START*3,sizeof(pal_rot));

	word startTime = ReadClock();
	while (ClockToSeconds(ReadClock()-startTime) < 25)
	{
		byte r1=pal_rot[0],g1=pal_rot[1],b1=pal_rot[2];
		byte *pal = pal_rot;
		for (int i=0;i<PAL_ANIM_COUNT-1;i++)
		{
			pal[0] = pal[3];
			pal[1] = pal[4];
			pal[2] = pal[5];
			pal += 3;
		}
		pal[0] = r1; pal[1] = g1; pal[2] = b1;

		pal = pal_rot;
		for (int i=0;i<PAL_ANIM_COUNT;i++)
		{
			outp(0x03C8,i+PAL_ANIM_START);
			outp(0x03C9,pal[0]); // R
			outp(0x03C9,pal[1]); // G
			outp(0x03C9,pal[2]); // B
			pal += 3;
		}

		if (AnyKeyPressed())
			return;

		while (!VerticalRetraceDone())
		{
			if (!midPlayDone)
				midPlayDone = bgmidTick(mid.params,mid.trackinfo,mid.eventscache,mid.playinfo) > 0;
		}
		while (!VerticalRefreshDone())
		{
			if (!midPlayDone)
				midPlayDone = bgmidTick(mid.params,mid.trackinfo,mid.eventscache,mid.playinfo) > 0;
		}
	}
}

int main(int argc,char **args)
{
	char *stdargs[] = {"SINTRO.EXE", "SIERRA.MID", "/mpu"};
	if (argc <= 1)
	{
		argc = 3;
		args = stdargs;
	}
	BITMAP256 *bmp = (BITMAP256*)malloc(sizeof(BITMAP256));
	midiinfo mid;

	bgmidInit(argc,args,&mid.params,&mid.trackinfo,&mid.eventscache,&mid.playinfo);

	// Switch to mode 13
	{
		REGS regs;
		regs.h.ah = 0x00; // Function 0, mode set
		regs.h.al = 0x13; // 256-colors at 320x200
		int86(0x10,&regs,&regs);
	}

	if (!bmp->LoadToVGA("SIERRA.BMP"))
	{
		free(bmp);
		return 1;
	}
	//bmp->FunkyPalette();

	bgmidBeginPlay(mid.params,mid.trackinfo,mid.eventscache,mid.playinfo);
	AnimateImagePalette(*bmp,mid);
	bgmidShutdown(&mid.params,&mid.trackinfo,&mid.eventscache,&mid.playinfo);

	// Switch back to mode 3
	{
		REGS regs;
		regs.h.ah = 0x00; // Function 0, mode set
		regs.h.al = 0x3; // Text mode
		int86(0x10,&regs,&regs);
	}

	free(bmp);

	puts("");
	puts("A shy reproduction of one of the intros for a company that once gave");
	puts("us so many wonderful games:");
	puts("     ͻ");
	puts("                       Sierra On-line                 ");
	puts("     ͼ");
	puts("");
	puts("Thank you for taking those few seconds to remember all the great");
	puts("people who worked hard to give us unforgettable stories and experiences!");
	puts("");
	puts("Credits:");
	puts("  Demo Programming: Wessam Bahnassi (@wbahnassi)");
	puts("  Music Player: Modified version of DOSMid (original code by Mateusz Viste)");
	puts("  Art: Captured from Space Quest 5 using the ever-awesome DOSBox");
	puts("  MIDI: Same as above!");
	puts("");
	puts("== 16-bit real-mode DOS C++ program compiled with OpenWatcom 1.9  ==");
	puts("== written in Windows 10 & Visual Studio 2017 on my Surface Pro 4 ==");
	puts("");
	return 0;
}